# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

import numpy as np
from typing import Tuple
import gym_ignition_models
from scenario import core
from scenario import gazebo as scenario
from ..common.utils import default_world_fixture as default_world

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)

# Panda PID gains
# https://github.com/mkrizmancic/franka_gazebo/blob/master/config/default.yaml
panda_pid_gains_1000Hz = {
    'panda_joint1': core.PID(50,    0,  20),
    'panda_joint2': core.PID(10000, 0, 500),
    'panda_joint3': core.PID(100,   0,  10),
    'panda_joint4': core.PID(1000,  0,  50),
    'panda_joint5': core.PID(100,   0,  10),
    'panda_joint6': core.PID(100,   0,  10),
    'panda_joint7': core.PID(10,    0.5, 0.1),
    'panda_finger_joint1': core.PID(100, 0, 50),
    'panda_finger_joint2': core.PID(100, 0, 50),
}


@pytest.mark.parametrize("default_world", [(1.0 / 1_000, 1.0, 1)], indirect=True)
def test_position_pid(default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

    # Get the simulator and the world
    gazebo, world = default_world

    # Insert a panda model
    panda_urdf = gym_ignition_models.get_model_file("panda")
    assert world.insert_model(panda_urdf)
    assert "panda" in world.model_names()

    # Show the GUI
    # import time
    # gazebo.gui()
    # gazebo.run(paused=True)
    # time.sleep(3)

    # Get the model and cast it to Gazebo
    panda = world.get_model("panda").to_gazebo()

    # Reset joint1 to its middle position
    joint1 = panda.get_joint("panda_joint1").to_gazebo()
    joint1_range = np.abs(joint1.position_limit().max - joint1.position_limit().min)
    joint1_middle = joint1.position_limit().min + joint1_range / 2
    assert joint1.reset_position(joint1_middle)

    # Reset joint6 to its middle position
    joint6 = panda.get_joint("panda_joint6").to_gazebo()
    joint6_range = np.abs(joint6.position_limit().max - joint6.position_limit().min)
    joint6_middle = joint6.position_limit().min + joint6_range / 2
    assert joint6.reset_position(joint6_middle)

    # Update the model state without stepping the physics
    assert gazebo.run(paused=True)

    # Set the controller period equal to the physics step (1000Hz)
    panda.set_controller_period(gazebo.step_size())

    # Check that we have gains for all joints
    assert set(panda.joint_names()) == set(panda_pid_gains_1000Hz.keys())

    # Set the PID gains
    for joint_name, pid in panda_pid_gains_1000Hz.items():
        assert panda.get_joint(joint_name).set_pid(pid=pid)

    # Switch to position control mode (it sets the current position as active target)
    assert panda.set_joint_control_mode(core.JointControlMode_position)
    assert panda.joint_position_targets() == pytest.approx(panda.joint_positions())

    # Just fight gravity for a while
    for _ in range(1_000):
        assert gazebo.run()

    # Check that it didn't move
    assert panda.joint_positions() == pytest.approx(panda.joint_position_targets(),
                                                    abs=np.deg2rad(1))

    # joint1 trajectory
    q0_joint1 = joint1.position()
    q_joint1 = (0.9 * joint1_range / 2 * np.sin(2 * np.pi * 0.33 * t)
                for t in np.arange(start=0, stop=10.0, step=gazebo.step_size()))

    # joint6 trajectory
    q0_joint6 = joint6.position()
    q_joint6 = (0.9 * joint6_range / 2 * np.sin(2 * np.pi * 0.33 * t)
                for t in np.arange(start=0, stop=10.0, step=gazebo.step_size()))

    for _ in range(5_000):

        # Generate the new references
        joint1_reference = q0_joint1 + next(q_joint1)
        joint6_reference = q0_joint6 + next(q_joint6)

        # Set the new references
        assert joint1.set_position_target(position=joint1_reference)
        assert joint6.set_position_target(position=joint6_reference)

        # Run the simulation
        assert gazebo.run()

        # Check that trajectory is being followed
        assert joint1.position() == pytest.approx(joint1_reference, abs=np.deg2rad(3))
        assert joint6.position() == pytest.approx(joint6_reference, abs=np.deg2rad(3))



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import Union
from gym_ignition_environments import tasks
from gym_ignition_environments.models import cartpole
from gym_ignition.randomizers import gazebo_env_randomizer
from gym_ignition.randomizers.gazebo_env_randomizer import MakeEnvCallable

# Tasks that are supported by this randomizer. Used for type hinting.
SupportedTasks = Union[tasks.cartpole_discrete_balancing.CartPoleDiscreteBalancing,
                       tasks.cartpole_continuous_swingup.CartPoleContinuousSwingup,
                       tasks.cartpole_continuous_balancing.CartPoleContinuousBalancing]


class CartpoleEnvNoRandomizations(gazebo_env_randomizer.GazeboEnvRandomizer):
    """
    Dummy environment randomizer for cartpole tasks.

    Check :py:class:`~gym_ignition_environments.randomizers.cartpole.CartpoleRandomizersMixin`
    for an example that randomizes the task, the physics, and the model.
    """

    def __init__(self, env: MakeEnvCallable):

        super().__init__(env=env)

    def randomize_task(self, task: SupportedTasks, **kwargs) -> None:
        """
        Prepare the scene for cartpole tasks. It simply removes the cartpole of the
        previous rollout and inserts a new one in the default state. Then, the active
        Task will reset the state of the cartpole depending on the implemented
        decision-making logic.
        """

        if "gazebo" not in kwargs:
            raise ValueError("gazebo kwarg not passed to the task randomizer")

        gazebo = kwargs["gazebo"]

        # Remove the model from the simulation
        if task.model_name is not None and task.model_name in task.world.model_names():

            if not task.world.to_gazebo().remove_model(task.model_name):
                raise RuntimeError("Failed to remove the cartpole from the world")

        # Execute a paused run to process model removal
        if not gazebo.run(paused=True):
            raise RuntimeError("Failed to execute a paused Gazebo run")

        # Insert a new cartpole model
        model = cartpole.CartPole(world=task.world)

        # Store the model name in the task
        task.model_name = model.name()

        # Execute a paused run to process model insertion
        if not gazebo.run(paused=True):
            raise RuntimeError("Failed to execute a paused Gazebo run")



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import time
import functools
from gym_ignition.utils import logger
from gym_ignition_environments import randomizers

# Set verbosity
logger.set_level(gym.logger.ERROR)
# logger.set_level(gym.logger.DEBUG)

# Available tasks
env_id = "CartPoleDiscreteBalancing-Gazebo-v0"
# env_id = "CartPoleContinuousBalancing-Gazebo-v0"
# env_id = "CartPoleContinuousSwingup-Gazebo-v0"


def make_env_from_id(env_id: str, **kwargs) -> gym.Env:
    import gym
    import gym_ignition_environments
    return gym.make(env_id, **kwargs)


# Create a partial function passing the environment id
make_env = functools.partial(make_env_from_id, env_id=env_id)

# Wrap the environment with the randomizer.
# This is a simple example no randomization are applied.
env = randomizers.cartpole_no_rand.CartpoleEnvNoRandomizations(env=make_env)

# Wrap the environment with the randomizer.
# This is a complex example that randomizes both the physics and the model.
# env = randomizers.cartpole.CartpoleEnvRandomizer(
#     env=make_env, seed=42, num_physics_rollouts=5)

# Enable the rendering
# env.render('human')

# Initialize the seed
env.seed(42)

for epoch in range(10):

    # Reset the environment
    observation = env.reset()

    # Initialize returned values
    done = False
    totalReward = 0

    while not done:

        # Execute a random action
        action = env.action_space.sample()
        observation, reward, done, _ = env.step(action)

        # Render the environment.
        # It is not required to call this in the loop if physics is not randomized.
        # env.render('human')

        # Accumulate the reward
        totalReward += reward

        # Print the observation
        msg = ""
        for value in observation:
            msg += "\t%.6f" % value
        logger.debug(msg)

    print(f"Reward episode #{epoch}: {totalReward}")

env.close()
time.sleep(5)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
import tempfile
import numpy as np
import pybullet_data
import pybullet as p
from gym_ignition.robots import sim
from pybullet_utils import bullet_client
from gym_ignition.robots import gazebo_robot
from gym_ignition.utils import resource_finder
from gym_ignition import gympp_bindings as bindings


class Simulator(abc.ABC):
    simulator = None
    simulator_name: str = ""

    @abc.abstractmethod
    def step(self):
        pass

    @abc.abstractmethod
    def close(self):
        pass


class Gazebo(Simulator):
    simulator_name = "gazebo"

    def __init__(self,
                 physics_rate: float,
                 iterations: int = int(1),
                 rtf=float(np.finfo(np.float32).max)):
        self.simulator = bindings.GazeboSimulator(iterations, rtf, physics_rate)
        assert self.simulator

        self.simulator.setVerbosity(4)

        empty_world = resource_finder.find_resource("DefaultEmptyWorld.world")
        ok_world = self.simulator.setupGazeboWorld(empty_world)
        assert ok_world

        ok_initialize = self.simulator.initialize()
        assert ok_initialize

    def step(self):
        assert self.simulator.initialized()

        ok_run = self.simulator.run()
        assert ok_run

    def close(self):
        ok_close = self.simulator.close()
        assert ok_close


class PyBullet(Simulator):
    simulator_name = "pybullet"

    def __init__(self, physics_rate: float):
        self.simulator = bullet_client.BulletClient(p.DIRECT)
        assert self.simulator

        self.simulator.setRealTimeSimulation(0)
        self.simulator.setGravity(0, 0, -9.81)
        self.simulator.setTimeStep(1.0 / physics_rate)

        self.simulator.setAdditionalSearchPath(pybullet_data.getDataPath())
        self.plane_id = self.simulator.loadURDF("plane_implicit.urdf")

    def step(self):
        self.simulator.stepSimulation()

    def close(self):
        pass


# TODO: we need urdf for the controllers
def get_pendulum(simulator: Simulator, **kwargs):
    if simulator.simulator_name == "gazebo":
        return sim.gazebo.pendulum.PendulumGazeboRobot(
            gazebo=simulator.simulator,
            **kwargs)
    elif simulator.simulator_name == "pybullet":
        return sim.pybullet.pendulum.PendulumPyBulletRobot(
            p=simulator.simulator,
            plane_id=simulator.plane_id,
            **kwargs)
    else:
        raise Exception(f"Simulator {simulator.simulator_name} not recognized")


def get_cartpole(simulator: Simulator, **kwargs):
    if simulator.simulator_name == "gazebo":
        return sim.gazebo.cartpole.CartPoleGazeboRobot(
            gazebo=simulator.simulator,
            **kwargs)
    elif simulator.simulator_name == "pybullet":
        return sim.pybullet.cartpole.CartPolePyBulletRobot(
            p=simulator.simulator,
            plane_id=simulator.plane_id,
            **kwargs)
    else:
        raise Exception(f"Simulator {simulator.simulator_name} not recognized")


class CubeGazeboRobot(gazebo_robot.GazeboRobot):

    def __init__(self, gazebo, initial_position: np.ndarray, model_file: str = None):

        if model_file is None:
            # Serialize the cube urdf
            handle, model_file = tempfile.mkstemp()
            with open(handle, 'w') as f:
                f.write(get_cube_urdf())

        # Initialize base class
        super().__init__(model_file=model_file,
                         gazebo=gazebo)

        ok_floating = self.set_as_floating_base(True)
        assert ok_floating, "Failed to set the robot as floating base"

        # Initial base position and orientation
        base_position = np.array(initial_position)
        base_orientation = np.array([1., 0., 0., 0.])
        ok_base_pose = self.set_initial_base_pose(base_position, base_orientation)
        assert ok_base_pose, "Failed to set base pose"

        # Insert the model in the simulation
        _ = self.gympp_robot


def get_cube_urdf() -> str:
    mass = 5.0
    edge = 0.2
    i = 1 / 12 * mass * (edge ** 2 + edge ** 2)
    cube_urdf = f"""
    <robot name="cube_robot" xmlns:xacro="http://www.ros.org/wiki/xacro">
        <link name="cube">
            <inertial>
              <origin rpy="0 0 0" xyz="0 0 0"/>
              <mass value="{mass}"/>
              <inertia ixx="{i}" ixy="0" ixz="0" iyy="{i}" iyz="0" izz="{i}"/>
            </inertial>
            <visual>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </visual>
            <collision>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </collision>
        </link>
    </robot>"""
    return cube_urdf



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

# Workaround for https://github.com/osrf/sdformat/issues/227.
# It has to be done before loading the bindings.
import gym_ignition_models
gym_ignition_models.setup_environment()

# Add IGN_GAZEBO_RESOURCE_PATH to the default search path
import os
from gym_ignition.utils import resource_finder
if "IGN_GAZEBO_RESOURCE_PATH" in os.environ:
    resource_finder.add_path_from_env_var("IGN_GAZEBO_RESOURCE_PATH")


def initialize_verbosity() -> None:

    import gym
    import scenario
    import gym_ignition.utils.logger

    if scenario.detect_install_mode() is scenario.InstallMode.Developer:
        gym_ignition.utils.logger.set_level(level=gym.logger.INFO,
                                            scenario_level=gym.logger.WARN)

    elif scenario.detect_install_mode() is scenario.InstallMode.User:
        gym_ignition.utils.logger.set_level(level=gym.logger.WARN,
                                            scenario_level=gym.logger.WARN)

    else:
        raise ValueError(scenario.detect_install_mode())


# Configure default verbosity
initialize_verbosity()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

import numpy as np
from scenario import core
from ..common import utils
import gym_ignition_models
from typing import Callable, Tuple
from scenario import gazebo as scenario
from scipy.spatial.transform import Rotation
from ..common.utils import default_world_fixture as default_world
from gym_ignition.utils.scenario import get_joint_positions_space

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


def to_wxyz(xyzw: np.ndarray) -> np.ndarray:

    if xyzw.shape != (4,):
        raise ValueError(xyzw)

    return xyzw[[3, 0, 1, 2]]


def to_xyzw(wxyz: np.ndarray) -> np.ndarray:

    if wxyz.shape != (4,):
        raise ValueError(wxyz)

    return wxyz[[1, 2, 3, 0]]


def to_matrix(quaternion: Tuple[float, float, float, float]) -> np.ndarray:

    quaternion_xyzw = to_xyzw(np.array(quaternion))
    return Rotation.from_quat(quaternion_xyzw).as_matrix()


def get_random_panda(gazebo: scenario.GazeboSimulator,
                     world: scenario.World) -> core.Model:

    panda_urdf = gym_ignition_models.get_model_file("panda")
    assert world.insert_model(panda_urdf)
    assert "panda" in world.model_names()

    panda = world.get_model("panda")

    joint_space = get_joint_positions_space(model=panda)
    joint_space.seed(10)

    q = joint_space.sample()
    dq = joint_space.np_random.uniform(low=-1.0, high=1.0, size=q.shape)

    assert panda.to_gazebo().reset_joint_positions(q.tolist())
    assert panda.to_gazebo().reset_joint_velocities(dq.tolist())

    assert gazebo.run(paused=True)
    return panda


def get_cube(gazebo: scenario.GazeboSimulator,
             world: scenario.World) -> core.Model:

    quaternion = to_wxyz(Rotation.from_euler('x', 45, degrees=True).as_quat())
    initial_pose = core.Pose([0, 0, 0.5], quaternion.tolist())

    cube_urdf = utils.get_cube_urdf()
    assert world.insert_model(cube_urdf, initial_pose)
    assert "cube_robot" in world.model_names()

    cube = world.get_model("cube_robot")

    assert cube.to_gazebo().reset_base_world_linear_velocity([0.1, -0.2, -0.3])
    assert cube.to_gazebo().reset_base_world_angular_velocity([-0.1, 2.0, 0.3])

    assert gazebo.run(paused=True)
    return cube


# 1. VELOCITY LINEAR
@pytest.mark.parametrize("default_world, get_model, link_name", [
                          ((1.0 / 10_000, 1.0, 1), get_random_panda, "panda_link7"),
                          ((1.0 / 10_000, 1.0, 1), get_cube, "cube"),
                          ],
                         indirect=["default_world"])
def test_linear_velocity(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World],
        get_model: Callable[[scenario.GazeboSimulator, scenario.World], core.Model],
        link_name: str):

    # Get the simulator and the world
    gazebo, world = default_world
    dt = gazebo.step_size()

    # Get the model
    model = get_model(gazebo, world)

    # Get the link.
    # If the link is the base link, get the data through the base methods.
    link = model.get_link(link_name)

    if link.name() != model.base_frame():

        position = link.position
        orientation = link.orientation
        body_linear_velocity = link.body_linear_velocity
        world_linear_velocity = link.world_linear_velocity

    else:

        position = model.base_position
        orientation = model.base_orientation
        body_linear_velocity = model.base_body_linear_velocity
        world_linear_velocity = model.base_world_linear_velocity

    # 0.5 seconds of simulation
    for _ in range(int(0.5 / dt)):

        position_old = np.array(position())
        assert gazebo.run()
        position_new = np.array(position())

        world_velocity = (position_new - position_old) / dt

        # Test the world velocity (MIXED)
        assert world_velocity == pytest.approx(world_linear_velocity(), abs=1E-2)

        # Eq 18
        # Test the BODY velocity
        W_R_L = to_matrix(orientation())
        body_velocity = W_R_L.transpose() @ np.array(world_linear_velocity())
        assert body_velocity == pytest.approx(body_linear_velocity())

    gazebo.close()


# 2. VELOCITY ANGULAR
@pytest.mark.parametrize("default_world, get_model, link_name", [
                          ((1.0 / 10_000, 1.0, 1), get_random_panda, "panda_link7"),
                          ((1.0 / 10_000, 1.0, 1), get_cube, "cube"),
                          ],
                         indirect=["default_world"])
def test_angular_velocity(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World],
        get_model: Callable[[scenario.GazeboSimulator, scenario.World], core.Model],
        link_name: str):

    # Get the simulator and the world
    gazebo, world = default_world
    dt = gazebo.step_size()

    # Get the model
    model = get_model(gazebo, world)

    # Get the link.
    # If the link is the base link, get the data through the base methods.
    link = model.get_link(link_name)

    if link.name() != model.base_frame():

        orientation = link.orientation
        body_angular_velocity = link.body_angular_velocity
        world_angular_velocity = link.world_angular_velocity

    else:

        orientation = model.base_orientation
        body_angular_velocity = model.base_body_angular_velocity
        world_angular_velocity = model.base_world_angular_velocity

    skew = lambda matrix: (matrix - matrix.transpose()) / 2
    vee = lambda matrix: [matrix[2, 1], matrix[0,2], matrix[1, 0]]

    for _ in range(int(0.5 / dt)):

        W_R_L_old = to_matrix(orientation())
        assert gazebo.run()
        W_R_L_new = to_matrix(orientation())

        dot_rotation_matrix = (W_R_L_new - W_R_L_old) / dt
        world_velocity = dot_rotation_matrix @ W_R_L_new.transpose()

        # Test the world velocity (MIXED)
        assert vee(skew(world_velocity)) == pytest.approx(world_angular_velocity(),
                                                          abs=0.005)

        # Test the BODY velocity
        body_velocity = W_R_L_new.transpose() @ dot_rotation_matrix
        assert vee(skew(body_velocity)) == pytest.approx(body_angular_velocity(),
                                                         abs=0.005)


# 3. ACCELERATION LINEAR
@pytest.mark.parametrize("default_world, get_model, link_name", [
                          ((1.0 / 10_000, 1.0, 1), get_random_panda, "panda_link7"),
                          # ((1.0 / 10_000, 1.0, 1), get_cube, "cube"),  # TODO
                          ],
                         indirect=["default_world"])
def test_linear_acceleration(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World],
        get_model: Callable[[scenario.GazeboSimulator, scenario.World], core.Model],
        link_name: str):

    # Get the simulator and the world
    gazebo, world = default_world
    dt = gazebo.step_size()

    # Get the model
    model = get_model(gazebo, world)

    # Get the link.
    # If the link is the base link, get the data through the base methods.
    link = model.get_link(link_name)

    if link.name() != model.base_frame():

        orientation = link.orientation
        world_linear_velocity = link.world_linear_velocity
        body_linear_acceleration = link.body_linear_acceleration
        world_linear_acceleration = link.world_linear_acceleration

    else:

        orientation = model.base_orientation
        world_linear_velocity = model.base_world_linear_velocity
        body_linear_acceleration = model.base_body_linear_acceleration
        world_linear_acceleration = model.base_word_linear_acceleration

    # 0.5 seconds of simulation
    for _ in range(int(0.5 / dt)):

        velocity_old = np.array(world_linear_velocity())
        assert gazebo.run()
        velocity_new = np.array(world_linear_velocity())

        world_acceleration = (velocity_new - velocity_old) / dt

        # By time to time there are steps where the acceleration becomes extremely high,
        # like 1000 m/s/s when the average is never exceeds 4 m/s/s.
        # We exclude those points. We should understand why this happens.
        if (np.array(world_linear_acceleration()) > 100.0).any():
            continue

        # Test the world acceleration (MIXED)
        assert world_linear_acceleration() == pytest.approx(world_acceleration, abs=0.5)

        # Test the BODY acceleration
        # Note: https://github.com/ignitionrobotics/ign-gazebo/issues/87
        W_R_L = to_matrix(orientation())
        body_acceleration = W_R_L.transpose() @ np.array(world_linear_acceleration())
        assert body_acceleration == pytest.approx(body_linear_acceleration())

    gazebo.close()


# 2. ACCELERATION ANGULAR
@pytest.mark.parametrize("default_world, get_model, link_name", [
                          ((1.0 / 10_000, 1.0, 1), get_random_panda, "panda_link7"),
                          # ((1.0 / 10_000, 1.0, 1), get_cube, "cube"),  # TODO
                          ],
                         indirect=["default_world"])
def test_angular_acceleration(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World],
        get_model: Callable[[scenario.GazeboSimulator, scenario.World], core.Model],
        link_name: str):

    # Get the simulator and the world
    gazebo, world = default_world
    dt = gazebo.step_size()

    # Get the model
    model = get_model(gazebo, world)

    # Get the link.
    # If the link is the base link, get the data through the base methods.
    link = model.get_link(link_name)

    if link.name() != model.base_frame():

        orientation = link.orientation
        world_angular_velocity = link.world_angular_velocity
        body_angular_acceleration = link.body_angular_acceleration
        world_angular_acceleration = link.world_angular_acceleration

    else:

        orientation = model.base_orientation
        world_angular_velocity = model.base_world_angular_velocity
        body_angular_acceleration = model.base_body_angular_acceleration
        world_angular_acceleration = model.base_word_angular_acceleration

    for _ in range(int(0.5 / dt)):

        world_velocity_old = np.array(world_angular_velocity())
        assert gazebo.run()
        world_velocity_new = np.array(world_angular_velocity())

        world_acceleration = (world_velocity_new - world_velocity_old) / dt

        # By time to time there are steps where the acceleration becomes extremely high,
        # like 1000 rad/s/s when the average is never exceeds 20 rad/s/s.
        # We exclude those points. We should understand why this happens.
        if (np.array(world_angular_acceleration()) > 100.0).any():
            continue

        # Test the world acceleration (MIXED)
        assert world_acceleration == pytest.approx(world_angular_acceleration(),
                                                   abs=0.2)

        # Note: https://github.com/ignitionrobotics/ign-gazebo/issues/87
        W_R_L = to_matrix(orientation())
        body_acceleration = W_R_L.transpose() @ np.array(world_angular_acceleration())
        assert body_acceleration == pytest.approx(body_angular_acceleration())



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import tempfile
import numpy as np
from . import utils


def test_external_force():
    # Get the simulator
    gazebo = utils.Gazebo(physics_rate=1000, iterations=1, rtf=100)

    # Create the first cube and insert it in the simulation
    cube = utils.CubeGazeboRobot(gazebo=gazebo.simulator,
                                 initial_position=np.array([0, 0, 1.0]))

    # Execute the first simulation step
    gazebo.step()

    # Get the position of the cube
    position, _ = cube.base_pose()
    assert position[0] == 0.0
    assert position[1] == 0.0

    # Apply a force for 1 second
    num_steps = 100
    f_x = 50.0
    f_y = - f_x / 10

    for _ in range(num_steps):
        ok_w = cube.apply_external_force(
            "cube", np.array([f_x, f_y, 0]), np.array([0.0, 0, 0]))
        assert ok_w

        # Step the simulation
        gazebo.step()

    # Get the position of the cube
    position, _ = cube.base_pose()
    print(position)
    assert position[0] > 0.0
    assert position[1] < 0.0
    assert np.allclose(-position[0], 10 * position[1])



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
import numpy as np
from enum import Enum, auto
from gym_ignition import rbd
import idyntree.bindings as idt
from dataclasses import dataclass
from typing import List, Dict, Optional, Union


class TargetType(Enum):

    POSITION = auto()
    ROTATION = auto()
    POSE = auto()


@dataclass
class TransformTargetData:

    position: np.ndarray
    quaternion: np.ndarray


@dataclass
class TargetData:

    type: TargetType
    weight: float
    data: Union[np.ndarray, TransformTargetData]


@dataclass
class IKSolution:

    joint_configuration: np.ndarray
    base_position: np.ndarray = np.array([0.0, 0.0, 0.0])
    base_quaternion: np.ndarray = np.array([1.0, 0.0, 0.0, 0.0])


class RotationParametrization(Enum):

    QUATERNION = auto()
    ROLL_PITCH_YAW = auto()

    def to_idyntree(self):

        if self.value == RotationParametrization.QUATERNION.value:
            return idt.InverseKinematicsRotationParametrizationQuaternion

        elif self.value == RotationParametrization.ROLL_PITCH_YAW.value:
            return idt.InverseKinematicsRotationParametrizationRollPitchYaw

        else:
            raise ValueError(self.value)


class TargetResolutionMode(Enum):

    TARGET_AS_CONSTRAINT_FULL = auto()
    TARGET_AS_CONSTRAINT_NONE = auto()
    TARGET_AS_CONSTRAINT_POSITION = auto()
    TARGET_AS_CONSTRAINT_ROTATION = auto()

    def to_idyntree(self):

        if self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_FULL.value:
            return idt.InverseKinematicsTreatTargetAsConstraintFull

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE.value:
            return idt.InverseKinematicsTreatTargetAsConstraintNone

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_POSITION.value:
            return idt.InverseKinematicsTreatTargetAsConstraintPositionOnly

        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_ROTATION.value:
            return idt.InverseKinematicsTreatTargetAsConstraintRotationOnly

        else:
            raise ValueError(self.value)


class InverseKinematicsNLP:

    def __init__(self,
                 urdf_filename: str,
                 considered_joints: List[str] = None,
                 joint_serialization: List[str] = None) -> None:

        self._floating_base: bool = False
        self._base_frame: Optional[str] = None
        self._urdf_filename: str = urdf_filename
        self._targets_data: Dict[str, TargetData] = dict()
        self._ik: Optional[idt.InverseKinematics] = None

        self._considered_joints: List[str] = considered_joints
        self._joint_serialization: List[str] = joint_serialization

    # ======================
    # INITIALIZATION METHODS
    # ======================

    def initialize(self,
                   rotation_parametrization: RotationParametrization =
                   RotationParametrization.ROLL_PITCH_YAW,
                   target_mode: TargetResolutionMode =
                   TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE,
                   cost_tolerance: float = 1E-08,
                   constraints_tolerance: float = 1E-4,
                   max_iterations: int = 1000,
                   base_frame: str = None,
                   floating_base: bool = False,
                   verbosity: int = 1) -> None:

        # Create the IK object
        self._ik = idt.InverseKinematics()

        # Load the URDF model and get the model loader object.
        # We create the full model with all the joints specified in joint_serialization.
        model_loader: idt.ModelLoader = self._get_model_loader(
            urdf=self._urdf_filename,
            joint_serialization=self._joint_serialization)

        # Get the model
        model: idt.Model = model_loader.model()

        # If all joints are enabled, get the list of joint names in order to know the
        # serialization of the full IK solution
        if self._joint_serialization is None:
            self._joint_serialization = []

            for joint_idx in range(model.getNrOfJoints()):
                self._joint_serialization.append(model.getJointName(joint_idx))

        # Configure the considered joints for the optimization.
        # If not specified, use the serialization of the full solution.
        if self._considered_joints is None:
            self._considered_joints = self._joint_serialization

        # Set the model in the IK specifying the considered joints
        if not self._ik.setModel(model, self._considered_joints):
            raise RuntimeError("Failed to set the model in the IK object")

        # Configure IK
        self._ik.setVerbosity(verbosity)
        self._ik.setMaxIterations(max_iterations)
        self._ik.setCostTolerance(cost_tolerance)
        self._ik.setConstraintsTolerance(constraints_tolerance)
        self._ik.setDefaultTargetResolutionMode(target_mode.to_idyntree())
        self._ik.setRotationParametrization(rotation_parametrization.to_idyntree())

        # Optionally change the base frame
        if base_frame is not None:
            # Store the frame of the base link
            self._base_frame = base_frame

            # Change the base frame
            if not self._ik.setFloatingBaseOnFrameNamed(base_frame):
                raise RuntimeError("Failed to change floating base frame")
        else:
            self._base_frame = model.getLinkName(model.getDefaultBaseLink())

        # Store whether the IK optimize a fixed or floating base robot
        self._floating_base = floating_base

        if not self._floating_base:

            # Add a frame constraint for the base
            self.add_frame_transform_constraint(frame_name=self._base_frame,
                                                position=np.array([0.0, 0, 0.0]),
                                                quaternion=np.array([1.0, 0, 0, 0]))

    def set_current_robot_configuration(self,
                                        base_position: np.ndarray,
                                        base_quaternion: np.ndarray,
                                        joint_configuration: np.ndarray) -> None:

        if joint_configuration.size != len(self._joint_serialization):
            raise ValueError(joint_configuration)

        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=base_position,
            quaternion=base_quaternion)

        q = rbd.idyntree.numpy.FromNumPy.to_idyntree_dyn_vector(array=joint_configuration)

        if not self._ik.setCurrentRobotConfiguration(H, q):
            raise RuntimeError("Failed to set the current robot configuration")

        if not self._floating_base:

            self.update_frame_transform_constraint(frame_name=self._base_frame,
                                                   position=base_position,
                                                   quaternion=base_quaternion)

    def set_current_joint_configuration(self,
                                        joint_name: str,
                                        configuration: float) -> None:

        if joint_name not in self._joint_serialization:
            raise ValueError(joint_name)

        if not self._ik.setJointConfiguration(jointName=joint_name,
                                              jointConfiguration=configuration):
            raise RuntimeError(f"Failed to set the configuration of joint '{joint_name}'")

    # ==============
    # TARGET METHODS
    # ==============

    def add_target(self,
                   frame_name: str,
                   target_type: TargetType,
                   weight: float = 1.0,
                   as_constraint: bool = False) -> None:

        if target_type == TargetType.ROTATION:
            # Add the target
            ok_target = self._ik.addRotationTarget(frame_name,
                                                   idt.Rotation.Identity(),
                                                   weight)

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.ROTATION,
                                                        weight=weight,
                                                        data=np.array([1.0, 0, 0, 0]))

        elif target_type == TargetType.POSITION:
            # Add the target
            ok_target = self._ik.addPositionTarget(frame_name,
                                                   idt.Position_Zero(),
                                                   weight)

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.POSITION,
                                                        weight=weight,
                                                        data=np.array([0.0, 0, 0]))

        elif target_type == TargetType.POSE:
            # Add the target
            ok_target = self._ik.addTarget(frame_name,
                                           idt.Transform.Identity(),
                                           weight,
                                           weight)

            # Create the transform target data
            target_data = TransformTargetData(position=np.array([0.0, 0, 0]),
                                              quaternion=np.array([1., 0, 0, 0]))

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.POSE,
                                                        weight=weight,
                                                        data=target_data)

        else:
            raise ValueError(target_type)

        if not ok_target:
            raise RuntimeError(f"Failed to add target for frame '{frame_name}'")

        if as_constraint:
            if target_type == TargetType.ROTATION:
                constraint = idt.InverseKinematicsTreatTargetAsConstraintRotationOnly
            elif target_type == TargetType.POSITION:
                constraint = idt.InverseKinematicsTreatTargetAsConstraintPositionOnly
            else:
                assert target_type == TargetType.POSE
                constraint = idt.InverseKinematicsTreatTargetAsConstraintFull

            if not self._ik.setTargetResolutionMode(frame_name, constraint):
                raise RuntimeError(f"Failed to set target '{frame_name}' as constraint")

    def add_com_target(self,
                       weight: float = 1.0,
                       as_constraint: bool = False,
                       constraint_tolerance: float = 1E-8) -> None:

        # Add the target
        self._ik.setCOMTarget(idt.Position_Zero(), weight)

        # Configure it either as target or constraint
        self._ik.setCOMAsConstraint(asConstraint=as_constraint)
        self._ik.setCOMAsConstraintTolerance(tolerance=constraint_tolerance)

        # Initialize the target data buffers
        assert "com" not in self._targets_data.keys()
        self._targets_data["com"] = TargetData(type=TargetType.POSITION,
                                               weight=weight,
                                               data=np.array([0.0, 0, 0]))

    def update_position_target(self, target_name: str, position: np.ndarray) -> None:

        if target_name not in self.get_active_target_names(
                target_type=TargetType.POSITION):

            raise ValueError(f"Failed to find a position target '{target_name}'")

        # Create the iDynTree position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Update the target inside IK
        if not self._ik.updateTarget(target_name, p):
            raise RuntimeError(f"Failed to update position of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.POSITION,
                                                     weight=weight,
                                                     data=position)

    def update_rotation_target(self, target_name: str, quaternion: np.ndarray) -> None:

        if target_name not in self.get_active_target_names(
                target_type=TargetType.ROTATION):

            raise ValueError(f"Failed to find a rotation target '{target_name}'")

        # Create the iDynTree rotation matrix
        R = rbd.idyntree.numpy.FromNumPy.to_idyntree_rotation(quaternion=quaternion)

        # Update the target inside IK
        if not self._ik.updateRotationTarget(target_name, R):
            raise RuntimeError(f"Failed to update rotation of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.ROTATION,
                                                     weight=weight,
                                                     data=quaternion)

    def update_transform_target(self,
                                target_name: str,
                                position: np.ndarray,
                                quaternion: np.ndarray) -> None:

        if target_name not in self.get_active_target_names(
                target_type=TargetType.POSE):

            raise ValueError(f"Failed to find a transform target '{target_name}'")

        # Create the iDynTree transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(position=position,
                                                               quaternion=quaternion)

        # Update the target inside IK
        if not self._ik.updateTarget(target_name, H):
            raise RuntimeError(f"Failed to update transform of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Create the transform target data
        transform_data = TransformTargetData(position=position, quaternion=quaternion)

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.POSE,
                                                     weight=weight,
                                                     data=transform_data)

    def update_com_target(self,
                          position: np.ndarray) -> None:

        if not self._ik.isCOMTargetActive():
            raise RuntimeError("Constraint on CoM not active")

        # Create the iDynTree position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Update the target inside IK
        self._ik.setCOMTarget(p, self._targets_data["com"].weight)

        # Update the target data
        self._targets_data["com"] = TargetData(type=TargetType.POSITION,
                                               weight=self._targets_data["com"].weight,
                                               data=position)

    def deactivate_com_target(self) -> None:

        if not self._ik.isCOMTargetActive():
            raise RuntimeError("Constraint on CoM not active")

        self._ik.deactivateCOMTarget()

    # =============
    # FRAME METHODS
    # =============

    def add_frame_transform_constraint(self,
                                       frame_name: str,
                                       position: np.ndarray,
                                       quaternion: np.ndarray) -> None:

        # Create the transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(position=position,
                                                               quaternion=quaternion)

        # Add the target
        if not self._ik.addFrameConstraint(frame_name, H):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_position_constraint(self,
                                      frame_name: str,
                                      position: np.ndarray) -> None:

        # Create the position
        p = rbd.idyntree.numpy.FromNumPy.to_idyntree_position(position=position)

        # Add the constraint
        if not self._ik.addFramePositionConstraint(frame_name, p):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_rotation_constraint(self,
                                      frame_name: str,
                                      quaternion: np.ndarray) -> None:

        # Create the position
        R = rbd.idyntree.numpy.FromNumPy.to_idyntree_rotation(quaternion=quaternion)

        # Add the constraint
        if not self._ik.addFrameRotationConstraint(frame_name, R):
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def update_frame_transform_constraint(self,
                                          frame_name: str,
                                          position: np.ndarray,
                                          quaternion: np.ndarray) -> None:

        if not self._ik.isFrameConstraintActive(frame_name):
            raise RuntimeError(f"Constraint on frame '{frame_name}' not active")

        # Create the transform
        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(position=position,
                                                               quaternion=quaternion)

        if not self._ik.activateFrameConstraint(frame_name, H):
            raise RuntimeError(f"Failed to update constraint on frame '{frame_name}'")

    def deactivate_frame_constraint(self, frame_name: str) -> None:

        if not self._ik.isFrameConstraintActive(frame_name):
            raise RuntimeError(f"Constraint on frame '{frame_name}' not active")

        if not self._ik.deactivateFrameConstraint(frame_name):
            raise RuntimeError(f"Failed to deactivate constraint on frame '{frame_name}'")

    # ===================
    # IK SOLUTION METHODS
    # ===================

    def solve(self) -> None:

        if not self._ik.solve():
            raise RuntimeError("Failed to solve IK")

        # Initialize next solver call
        self._warm_start_with_last_solution()

    def warm_start_from(self,
                        full_solution: IKSolution = None,
                        reduced_solution: IKSolution = None) -> None:

        if full_solution is None and reduced_solution is None or \
                full_solution is not None and reduced_solution is not None:
            raise ValueError("You have to specify either a full or a reduced solution")

        if reduced_solution is not None and\
                reduced_solution.joint_configuration.size != len(self._considered_joints):
            raise RuntimeError(
                "The joint configuration does not match the number of considered joints")

        if full_solution is not None and \
                full_solution.joint_configuration.size != len(self._joint_serialization):
            raise RuntimeError(
                "The joint configuration does not match the number of model joints")

        solution = reduced_solution if reduced_solution is not None else full_solution

        H = rbd.idyntree.numpy.FromNumPy.to_idyntree_transform(
            position=solution.base_position,
            quaternion=solution.base_quaternion)

        q = rbd.idyntree.numpy.FromNumPy.to_idyntree_dyn_vector(
            array=solution.joint_configuration)

        # Warm start the solver
        if not self._ik.setFullJointsInitialCondition(H, q):
            raise RuntimeError("Failed to warm start the IK solver")

    # =======
    # GETTERS
    # =======

    def get_base_frame(self) -> str:

        return self._base_frame

    def get_available_target_names(self) -> List[str]:

        # Get the reduced model
        model = self._ik.reducedModel()

        # Note that also frames (modeled as fake links) are available targets
        link_names = []

        for link_index in range(model.getNrOfLinks()):
            link_names.append(model.getLinkName(link_index))

        return link_names

    def get_active_target_names(self, target_type: TargetType = None) -> List[str]:

        if target_type is None:
            return list(self._targets_data.keys())

        else:
            return [name for name, value in self._targets_data.items()
                    if value.type == target_type ]

    def get_target_data(self, target_name: str) -> TargetData:

        return self._targets_data[target_name]

    def get_full_solution(self) -> IKSolution:

        # Initialize buffers
        base_transform = idt.Transform.Identity()
        joint_positions = idt.VectorDynSize(self._ik.fullModel().getNrOfJoints())
        assert len(joint_positions) == len(self._joint_serialization)

        # Get the solution
        self._ik.getFullJointsSolution(base_transform, joint_positions)

        # Convert to numpy objects
        joint_positions = joint_positions.toNumPy()
        base_position = base_transform.getPosition().toNumPy()
        base_quaternion = base_transform.getRotation().asQuaternion().toNumPy()

        return IKSolution(base_position=base_position,
                          base_quaternion=base_quaternion,
                          joint_configuration=joint_positions)

    def get_reduced_solution(self) -> IKSolution:

        # Initialize buffers
        base_transform = idt.Transform.Identity()
        joint_positions = idt.VectorDynSize(self._ik.reducedModel().getNrOfJoints())
        assert len(joint_positions) == len(self._considered_joints)

        # Get the solution
        self._ik.getReducedSolution(base_transform, joint_positions)

        # Convert to numpy objects
        joint_positions = joint_positions.toNumPy()
        base_position = base_transform.getPosition().toNumPy()
        base_quaternion = base_transform.getRotation().asQuaternion().toNumPy()

        return IKSolution(base_position=base_position,
                          base_quaternion=base_quaternion,
                          joint_configuration=joint_positions)

    # ===============
    # PRIVATE METHODS
    # ===============

    @staticmethod
    def _get_model_loader(urdf: str,
                          joint_serialization: List[str] = None) \
            -> idt.ModelLoader:

        # Get the model loader
        model_loader = idt.ModelLoader()

        if not os.path.exists(urdf):
            raise FileNotFoundError(urdf)

        # Load the model
        if joint_serialization:
            ok_load = model_loader.loadReducedModelFromFile(urdf, joint_serialization)
        else:
            ok_load = model_loader.loadModelFromFile(urdf)

        if not ok_load:
            raise RuntimeError("Failed to load model")

        # Due to some SWIG internal, returning the model contained by the loader
        # does not work as expected
        return model_loader

    def _warm_start_with_last_solution(self) -> None:

        last_solution = self.get_full_solution()
        self.warm_start_from(full_solution=last_solution)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

import numpy as np
from typing import Tuple
from scenario import core
from ..common import utils
import gym_ignition_models
from scenario import gazebo as scenario
from ..common.utils import gazebo_fixture as gazebo
from ..common.utils import default_world_fixture as default_world

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


def get_model(gazebo: scenario.GazeboSimulator,
              gym_ignition_models_name: str) -> scenario.Model:

    # Get the world and cast it to a Gazebo world
    world = gazebo.get_world().to_gazebo()

    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    model_urdf = gym_ignition_models.get_model_file(gym_ignition_models_name)

    assert world.insert_model(model_urdf,
                              core.Pose_identity(),
                              gym_ignition_models_name)

    # Get the model and cast it to a Gazebo model
    model = world.get_model(gym_ignition_models_name).to_gazebo()
    assert model.id() != 0
    assert model.valid()

    return model


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_model_core_api(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    gym_ignition_model_name = "cartpole"
    model = get_model(gazebo, gym_ignition_model_name)

    assert model.id() != 0
    assert model.valid()
    assert model.name() == gym_ignition_model_name

    assert len(model.link_names()) == model.nr_of_links()
    assert len(model.joint_names()) == model.nr_of_joints()

    assert model.set_controller_period(0.42)
    assert model.controller_period() == 0.42


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_model_joints(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    gym_ignition_model_name = "panda"
    model = get_model(gazebo, gym_ignition_model_name)

    q = model.joint_positions()
    assert pytest.approx(q) == [0.0] * model.dofs()

    dq = model.joint_velocities()
    assert pytest.approx(dq) == [0.0] * model.dofs()

    assert model.reset_joint_positions([0.05] * model.dofs())
    assert model.joint_positions() == pytest.approx([0.0] * model.dofs())
    gazebo.run(paused=True)
    assert model.joint_positions() == pytest.approx([0.05] * model.dofs())
    assert model.joint_velocities() == pytest.approx([0.0] * model.dofs())

    gazebo.run(paused=False)
    assert model.joint_velocities() != pytest.approx([0.0] * model.dofs())

    assert model.reset_joint_velocities([-0.1] * model.dofs())
    assert model.joint_velocities() != pytest.approx([-0.1] * model.dofs())
    gazebo.run(paused=True)
    assert model.joint_velocities() == pytest.approx([-0.1] * model.dofs())

    assert model.reset_joint_positions([0.0] * model.dofs())
    assert model.reset_joint_velocities([0.0] * model.dofs())

    joint_subset = model.joint_names()[0:4]
    assert model.reset_joint_positions([-0.4] * len(joint_subset), joint_subset)
    assert model.reset_joint_velocities([3.0] * len(joint_subset), joint_subset)
    gazebo.run(paused=True)
    assert model.joint_positions(joint_subset) == \
        pytest.approx([-0.4] * len(joint_subset))
    assert model.joint_velocities(joint_subset) == \
        pytest.approx([3.0] * len(joint_subset))
    assert model.joint_positions() == pytest.approx(
        [-0.4] * len(joint_subset) + [0.0] * (model.dofs() - len(joint_subset)))
    assert model.joint_velocities() == pytest.approx(
        [3.0] * len(joint_subset) + [0.0] * (model.dofs() - len(joint_subset)))


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_model_base_pose(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    gym_ignition_model_name = "pendulum"
    model = get_model(gazebo, gym_ignition_model_name)
    assert gym_ignition_model_name in gazebo.get_world().model_names()

    assert model.base_frame() == "support"
    # assert model.set_base_frame("support")  # TODO: Not yet supported
    # assert model.set_base_frame("pendulum")  # TODO: Not yet supported

    # Check that the pose is the identical
    gazebo.run(paused=True)
    assert model.base_position() == pytest.approx([0, 0, 0])
    assert model.base_orientation() == pytest.approx([1, 0, 0, 0])

    # Reset the base pose
    new_base_pose = dict(position=[5, 5, 0], orientation=[0, 1, 0, 0])
    assert model.reset_base_pose(new_base_pose['position'], new_base_pose['orientation'])

    # Before stepping the simulation the pose should be the initial one
    assert model.base_position() == pytest.approx([0, 0, 0])
    assert model.base_orientation() == pytest.approx([1, 0, 0, 0])

    # Step the simulator and check that the pose changes
    gazebo.run(paused=True)
    assert model.base_position() == pytest.approx(new_base_pose['position'])
    assert model.base_orientation() == pytest.approx(new_base_pose['orientation'])


@pytest.mark.parametrize("default_world",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_model_base_velocity(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

    # Get the simulator and the world
    gazebo, world = default_world

    model_urdf = gym_ignition_models.get_model_file("pendulum")

    # Insert the first pendulum
    model1_name = "pendulum1"
    model1_pose = core.Pose([0, 1.0, 1.0], [1.0, 0, 0, 0])
    assert world.insert_model(model_urdf, model1_pose, model1_name)
    model1 = world.get_model(model1_name).to_gazebo()
    assert model1.valid()
    assert model1_name in gazebo.get_world().model_names()
    assert model1.base_frame() == "support"

    # Insert the second pendulum
    model2_name = "pendulum2"
    model2_pose = core.Pose([0, -1.0, 1.0], [1.0, 0, 0, 0])
    assert world.insert_model(model_urdf, model2_pose, model2_name)
    model2 = world.get_model(model2_name).to_gazebo()
    assert model2.valid()
    assert model2_name in gazebo.get_world().model_names()
    assert model2.base_frame() == "support"

    # Target velocity
    lin_velocity = [0, 0, 5.0]
    ang_velocity = [0.1, 0.5, -3.0]

    # Reset both linear and angular of pendulum1 by calling the combined method
    assert model1.reset_base_world_velocity(lin_velocity, ang_velocity)
    assert model1.base_world_linear_velocity() == pytest.approx([0, 0, 0])
    assert model1.base_world_angular_velocity() == pytest.approx([0, 0, 0])

    # Reset both linear and angular of pendulum2 by calling the individual method
    assert model2.reset_base_world_linear_velocity(lin_velocity)
    assert model2.reset_base_world_angular_velocity(ang_velocity)
    assert model2.base_world_linear_velocity() == pytest.approx([0, 0, 0])
    assert model2.base_world_angular_velocity() == pytest.approx([0, 0, 0])

    # Run the simulation
    gazebo.run()

    # Check model1 velocity
    assert model1.base_world_linear_velocity() == pytest.approx(lin_velocity, abs=0.01)
    assert model1.base_world_angular_velocity() == pytest.approx(ang_velocity, abs=0.01)

    # Check model2 velocity
    assert model2.base_world_linear_velocity() == pytest.approx(lin_velocity, abs=0.01)
    assert model2.base_world_angular_velocity() == pytest.approx(ang_velocity, abs=0.01)

    # The velocity of the base link must be the same
    assert model1.get_link("support").world_linear_velocity() == \
           pytest.approx(lin_velocity, abs=0.01)
    assert model1.get_link("support").world_angular_velocity() == \
           pytest.approx(ang_velocity, abs=0.01)

    # The velocity of the base link must be the same
    assert model2.get_link("support").world_linear_velocity() == \
           pytest.approx(lin_velocity, abs=0.01)
    assert model2.get_link("support").world_angular_velocity() == \
           pytest.approx(ang_velocity, abs=0.01)


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_model_references(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    gym_ignition_model_name = "cartpole"
    model = get_model(gazebo, gym_ignition_model_name)
    assert gym_ignition_model_name in gazebo.get_world().model_names()

    assert model.set_joint_control_mode(core.JointControlMode_force)

    assert model.set_joint_position_targets([0.5, -3])
    assert model.joint_position_targets() == pytest.approx([0.5, -3])
    assert model.joint_position_targets(["pivot"]) == pytest.approx([-3])

    assert model.set_joint_velocity_targets([-0.1, 6])
    assert model.joint_velocity_targets() == pytest.approx([-0.1, 6])
    assert model.joint_velocity_targets(["pivot"]) == pytest.approx([6])

    assert model.set_joint_acceleration_targets([-0, 3.14])
    assert model.joint_acceleration_targets() == pytest.approx([-0, 3.14])
    assert model.joint_acceleration_targets(["pivot"]) == pytest.approx([3.14])

    assert model.set_joint_generalized_force_targets([20.1, -13])
    assert model.joint_generalized_force_targets() == pytest.approx([20.1, -13])
    assert model.joint_generalized_force_targets(["pivot"]) == pytest.approx([-13])

    assert model.set_base_pose_target((0, 0, 5), (0, 0, 0, 1.0))
    assert model.set_base_orientation_target((0, 0, 1.0, 0))
    assert model.base_position_target() == pytest.approx([0, 0, 5])
    assert model.base_orientation_target() == pytest.approx([0, 0, 1.0, 0])

    assert model.set_base_world_linear_velocity_target((1, 2, 3))
    assert model.set_base_world_angular_velocity_target((4, 5, 6))
    assert model.set_base_world_angular_acceleration_target((-1, -2, -3))
    assert model.base_world_linear_velocity_target() == pytest.approx([1, 2, 3])
    assert model.base_world_angular_velocity_target() == pytest.approx([4, 5, 6])
    assert model.base_world_angular_acceleration_target() == pytest.approx([-1, -2, -3])


# def test_model_contacts():
#     pass


@pytest.mark.parametrize("default_world", [(1.0 / 1_000, 1.0, 1)], indirect=True)
def test_history_of_joint_forces(
        default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

    # Get the simulator and the world
    gazebo, world = default_world

    # Insert a panda model
    panda_urdf = gym_ignition_models.get_model_file("panda")
    assert world.insert_model(panda_urdf)
    assert "panda" in world.model_names()

    # Get the model
    panda = world.get_model("panda")

    # Control the robot in Force
    assert panda.set_joint_control_mode(core.JointControlMode_force)

    # Enable the history for 3 steps
    assert panda.enable_history_of_applied_joint_forces(True, 3)

    # Initialize the torques applied in the first run
    torques = np.zeros(panda.dofs())

    # History of joint forces
    history_last_three_runs = np.zeros(panda.dofs() * 3)

    for _ in range(10):

        torques += 0.1
        assert panda.set_joint_generalized_force_targets(torques)

        gazebo.run()

        history_last_three_runs = np.concatenate((history_last_three_runs, torques))
        history_last_three_runs = history_last_three_runs[-panda.dofs() * 3:]

        assert panda.history_of_applied_joint_forces() == \
            pytest.approx(history_last_three_runs)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gym_ignition

import gym
from gym_ignition.utils.logger import set_level
from gym_ignition_environments import randomizers

# Set the verbosity
set_level(gym.logger.DEBUG)


def make_env(**kwargs) -> gym.Env:

    import gym
    import gym_ignition_environments
    return gym.make("CartPoleDiscreteBalancing-Gazebo-v0", **kwargs)


@pytest.mark.parametrize("num_physics_rollouts", [0, 2])
def test_reproducibility(num_physics_rollouts: int):

    env1 = randomizers.cartpole.CartpoleEnvRandomizer(
        env=make_env, num_physics_rollouts=num_physics_rollouts)

    env2 = randomizers.cartpole.CartpoleEnvRandomizer(
        env=make_env, num_physics_rollouts=num_physics_rollouts)

    assert env1 != env2

    # Seed the environment
    env1.seed(42)
    env2.seed(42)

    for _ in range(5):

        # Reset the environments
        observation1 = env1.reset()
        observation2 = env2.reset()
        assert observation1 == pytest.approx(observation2)

        # Initialize returned values
        done = False

        while not done:

            # Sample a random action
            action1 = env1.action_space.sample()
            action2 = env2.action_space.sample()
            assert action1 == pytest.approx(action2)

            # Step the environment
            observation1, reward1, done1, info1 = env1.step(action1)
            observation2, reward2, done2, info2 = env2.step(action2)

            assert done1 == pytest.approx(done2)
            assert info1 == pytest.approx(info2)
            assert reward1 == pytest.approx(reward2)
            assert observation1 == pytest.approx(observation2)

            done = done1

    env1.close()
    env2.close()


# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy
from . import tasks
from . import models
from . import randomizers
from gym.envs.registration import register


max_float = float(numpy.finfo(numpy.float32).max)

register(
    id='Pendulum-Gazebo-v0',
    entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
    max_episode_steps=5000,
    kwargs={'task_cls': tasks.pendulum_swingup.PendulumSwingUp,
            'agent_rate': 1000,
            'physics_rate': 1000,
            'real_time_factor': max_float,
            })

register(
    id='CartPoleDiscreteBalancing-Gazebo-v0',
    entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
    max_episode_steps=5000,
    kwargs={'task_cls': tasks.cartpole_discrete_balancing.CartPoleDiscreteBalancing,
            'agent_rate': 1000,
            'physics_rate': 1000,
            'real_time_factor': max_float,
            })

register(
    id='CartPoleContinuousBalancing-Gazebo-v0',
    entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
    max_episode_steps=5000,
    kwargs={'task_cls': tasks.cartpole_continuous_balancing.CartPoleContinuousBalancing,
            'agent_rate': 1000,
            'physics_rate': 1000,
            'real_time_factor': max_float,
            })

register(
    id='CartPoleContinuousSwingup-Gazebo-v0',
    entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
    max_episode_steps=5000,
    kwargs={'task_cls': tasks.cartpole_continuous_swingup.CartPoleContinuousSwingup,
            'agent_rate': 1000,
            'physics_rate': 1000,
            'real_time_factor': max_float,
            })



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import time
import pytest
import itertools
import numpy as np
from . import utils
from typing import List, Tuple
from gym_ignition.utils import logger
from gym_ignition.base.robot.robot_joints import JointControlMode


def almost_equal(first, second, epsilon=None) -> bool:
    if not epsilon:
        epsilon = first * 0.05

    res = np.abs(first - second) <= epsilon

    if not res:
        print("----------------")
        print("ASSERTION FAILED")
        print("----------------")
        print(f"#1={first}")
        print(f"#2={second}")
        print("Error={}".format(np.abs(first-second)))
        print(f"Tolerance={epsilon}")
        print("----------------")

    return res


def template_test(rtf: float,
                  physics_rate: float,
                  agent_rate: float,
                  ) -> None:

    # Get the simulator
    iterations = int(physics_rate / agent_rate)
    gazebo = utils.Gazebo(physics_rate=physics_rate, iterations=iterations, rtf=rtf)

    # Get the cartpole robot
    robot = utils.get_cartpole(gazebo)
    assert robot.valid(), "The robot object is not valid"

    # Configure the cartpole
    ok_cm = robot.set_joint_control_mode("linear", JointControlMode.POSITION)
    assert ok_cm, "Failed to set the control mode"

    # Set the joint controller period
    robot.set_dt(0.001 if physics_rate < 1000 else physics_rate)

    # Trajectory specifications
    trajectory_dt = 1 / agent_rate
    tot_simulated_seconds = 2

    # Generate the cart trajectory. This is analogous of setting an action containing
    # the cart references, hence it is related to the agent rate.
    cart_ref = np.fromiter(
        (0.2 * np.sin(2 * np.pi * 0.5 * t) for t in np.arange(0, tot_simulated_seconds,
                                                              trajectory_dt)),
        dtype=np.float)

    avg_time_per_step = 0.0
    start = time.time()

    logger.debug(f"Simulating {cart_ref.size} steps")
    for i, ref in enumerate(cart_ref):
        # Set the references
        ok_ref = robot.set_joint_position("linear", ref)
        assert ok_ref, "Failed to set joint references"

        # Step the simulator
        now = time.time()
        gazebo.step()
        this_step = time.time() - now
        avg_time_per_step = avg_time_per_step + 0.1 * (this_step - avg_time_per_step)

    stop = time.time()
    elapsed_time = stop - start

    # Compute useful quantities
    number_of_actions = tot_simulated_seconds / trajectory_dt
    physics_iterations_per_run = physics_rate / agent_rate
    simulation_dt = 1 / (physics_rate * rtf)

    assert number_of_actions == cart_ref.size

    print()
    print(f"Elapsed time = {elapsed_time}")
    print(f"Average step time = {avg_time_per_step}")
    print("Number of actions: = {}".format(tot_simulated_seconds / trajectory_dt))
    print(f"Physics iteration per simulator step = {physics_iterations_per_run}")
    print()

    # Check if the average time per step matches what expected
    assert almost_equal(first=avg_time_per_step,
                        second=simulation_dt * physics_iterations_per_run,
                        epsilon=avg_time_per_step * 0.5)

    # Check if the total time of the simulation matched what expected
    assert almost_equal(first=elapsed_time,
                        second=number_of_actions * simulation_dt * physics_iterations_per_run,
                        epsilon=elapsed_time * 0.5)

    # Terminate simulation
    gazebo.close()


def create_test_matrix() -> List[Tuple[float, float, float]]:
    rtf_list = [0.5, 1, 2, 5, 10]
    agent_rate_list = [100, 500, 1000]
    physics_rate_list = [100, 500, 1000]

    matrix = list()

    for agent_rate in agent_rate_list:
        # Compute all the combinations of simulation rate and physics rate
        combinations = list(itertools.product(rtf_list, physics_rate_list))

        for rtf, physics_rate in combinations:
            # Remove some combination too demanding for a single process
            if physics_rate * rtf >= 2000:
                continue

            # This case is not allowed
            if agent_rate > physics_rate:
                continue

            matrix.append((rtf, agent_rate, physics_rate))

    print(matrix)
    return matrix


@pytest.mark.parametrize("rtf, agent_rate, physics_rate", create_test_matrix())
def test_rates(rtf: float, agent_rate: float, physics_rate: float):

    print("========")
    print("Testing:")
    print("========")
    print(f"RTF = {rtf}")
    print(f"Agent rate = {agent_rate}")
    print(f"Physics rate = {physics_rate}")

    template_test(rtf, physics_rate, agent_rate)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

import numpy as np
from scenario import core
from ..common import utils
import gym_ignition_models
from typing import Callable
from gym_ignition.utils import misc
from scenario import gazebo as scenario
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


def get_cube_urdf_string_double_collision() -> str:

    mass = 5.0
    edge = 0.2
    i = 1 / 12 * mass * (edge ** 2 + edge ** 2)
    cube_urdf = f"""
    <robot name="cube_robot" xmlns:xacro="http://www.ros.org/wiki/xacro">
        <link name="cube">
            <inertial>
              <origin rpy="0 0 0" xyz="0 0 0"/>
              <mass value="{mass}"/>
              <inertia ixx="{i}" ixy="0" ixz="0" iyy="{i}" iyz="0" izz="{i}"/>
            </inertial>
            <visual>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </visual>
            <collision>
              <geometry>
                <box size="{edge} {edge / 2} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 -{edge / 4} 0"/>
            </collision>
            <collision>
              <geometry>
                <box size="{edge} {edge / 2} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 {edge / 4} 0"/>
            </collision>
        </link>
    </robot>"""

    return cube_urdf


@pytest.mark.parametrize("gazebo, get_model_str",
                         [((0.001, 1.0, 1), utils.get_cube_urdf_string),
                          ((0.001, 1.0, 1), get_cube_urdf_string_double_collision)],
                         indirect=["gazebo"],
                         ids=utils.id_gazebo_fn)
def test_cube_contact(gazebo: scenario.GazeboSimulator,
                      get_model_str: Callable):

    assert gazebo.initialize()
    world = gazebo.get_world().to_gazebo()

    # Insert the Physics system
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    # Insert the ground plane
    assert world.insert_model(gym_ignition_models.get_model_file("ground_plane"))
    assert len(world.model_names()) == 1

    # Insert the cube
    cube_urdf = misc.string_to_file(get_model_str())
    assert world.insert_model(cube_urdf,
                              core.Pose([0, 0, 0.15], [1., 0, 0, 0]),
                              "cube")
    assert len(world.model_names()) == 2

    # Get the cube
    cube = world.get_model("cube")

    # Enable contact detection
    assert not cube.contacts_enabled()
    assert cube.enable_contacts(enable=True)
    assert cube.contacts_enabled()

    # The cube was inserted floating in the air with a 5cm gap wrt to ground.
    # There should be no contacts.
    gazebo.run(paused=True)
    assert not cube.get_link("cube").in_contact()
    assert len(cube.contacts()) == 0

    # Make the cube fall for 150ms
    for _ in range(150):
        gazebo.run()

    # There should be only one contact, between ground and the cube
    assert cube.get_link("cube").in_contact()
    assert len(cube.contacts()) == 1

    # Get the contact
    contact_with_ground = cube.contacts()[0]

    assert contact_with_ground.body_a == "cube::cube"
    assert contact_with_ground.body_b == "ground_plane::link"

    # Check contact points
    for point in contact_with_ground.points:

        assert point.normal == pytest.approx([0, 0, 1])

    # Check that the contact force matches the weight of the cube
    z_forces = [point.force[2] for point in contact_with_ground.points]
    assert np.sum(z_forces) == pytest.approx(-5 * world.gravity()[2], abs=0.1)

    # Forces of all contact points are combined by the following method
    assert cube.get_link("cube").contact_wrench() == \
        pytest.approx([0, 0, np.sum(z_forces), 0, 0, 0])


@pytest.mark.parametrize("gazebo, get_model_str",
                         [((0.001, 1.0, 1), utils.get_cube_urdf_string),
                          ((0.001, 1.0, 1), get_cube_urdf_string_double_collision)],
                         indirect=["gazebo"],
                         ids=utils.id_gazebo_fn)
def test_cube_multiple_contacts(gazebo: scenario.GazeboSimulator,
                                get_model_str: Callable):

    assert gazebo.initialize()
    world = gazebo.get_world().to_gazebo()

    # Insert the Physics system
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    # Insert the ground plane
    assert world.insert_model(gym_ignition_models.get_model_file("ground_plane"))
    assert len(world.model_names()) == 1

    # Insert two cubes side to side with a 10cm gap
    cube_urdf = misc.string_to_file(get_model_str())
    assert world.insert_model(cube_urdf,
                              core.Pose([0, -0.15, 0.101], [1., 0, 0, 0]),
                              "cube1")
    assert world.insert_model(cube_urdf,
                              core.Pose([0, 0.15, 0.101], [1., 0, 0, 0]),
                              "cube2")
    assert len(world.model_names()) == 3

    # Get the cubes
    cube1 = world.get_model("cube1")
    cube2 = world.get_model("cube2")

    # Enable contact detection
    assert not cube1.contacts_enabled()
    assert cube1.enable_contacts(enable=True)
    assert cube1.contacts_enabled()

    # Enable contact detection
    assert not cube2.contacts_enabled()
    assert cube2.enable_contacts(enable=True)
    assert cube2.contacts_enabled()

    # The cubes were inserted floating in the air with a 1mm gap wrt to ground.
    # There should be no contacts.
    gazebo.run(paused=True)
    assert not cube1.get_link("cube").in_contact()
    assert not cube2.get_link("cube").in_contact()
    assert len(cube1.contacts()) == 0
    assert len(cube2.contacts()) == 0

    # Make the cubes fall for 50ms
    for _ in range(50):
        gazebo.run()

    assert cube1.get_link("cube").in_contact()
    assert cube2.get_link("cube").in_contact()
    assert len(cube1.contacts()) == 1
    assert len(cube2.contacts()) == 1

    # Now we make another cube fall above the gap. It will touch both cubes.
    assert world.insert_model(cube_urdf,
                              core.Pose([0, 0, 0.301], [1., 0, 0, 0]),
                              "cube3")
    assert len(world.model_names()) == 4

    cube3 = world.get_model("cube3")
    assert not cube3.contacts_enabled()
    assert cube3.enable_contacts(enable=True)
    assert cube3.contacts_enabled()

    gazebo.run(paused=True)
    assert not cube3.get_link("cube").in_contact()
    assert len(cube3.contacts()) == 0

    # Make the cube fall for 50ms
    for _ in range(50):
        gazebo.run()

    # There will be two contacts, respectively with cube1 and cube2.
    # Contacts are related to the link, not the collision elements. In the case of two
    # collisions elements associated to the same link, contacts are merged together.
    assert cube3.get_link("cube").in_contact()
    assert len(cube3.contacts()) == 2

    contact1 = cube3.contacts()[0]
    contact2 = cube3.contacts()[1]

    assert contact1.body_a == "cube3::cube"
    assert contact2.body_a == "cube3::cube"

    assert contact1.body_b == "cube1::cube"
    assert contact2.body_b == "cube2::cube"

    # Calculate the total contact wrench of cube3
    assert cube3.get_link("cube").contact_wrench() == pytest.approx([0, 0, 50, 0, 0, 0],
                                                                    abs=1.1)

    # Calculate the total contact force of the cubes below.
    # They will have 1.5 their weight from below and -0.5 from above.
    assert cube1.get_link("cube").contact_wrench()[2] == pytest.approx(50, abs=1.1)
    assert cube1.get_link("cube").contact_wrench()[2] == pytest.approx(50, abs=1.1)

    # Check the normals and the sign of the forces
    for contact in cube2.contacts():
        if contact.body_b == "cube3::cube":
            for point in contact.points:
                assert point.force[2] < 0
                assert point.normal == pytest.approx([0, 0, -1], abs=0.001)
        if contact.body_b == "ground_plane::link":
            for point in contact.points:
                assert point.force[2] > 0
                assert point.normal == pytest.approx([0, 0, 1], abs=0.001)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import warnings
import contextlib
from gym import logger
from gym.utils import colorize
from gym.logger import debug, info, error


def custom_formatwarning(msg, *args, **kwargs):
    """
    Custom format that overrides :py:func:`warnings.formatwarning`.
    """

    if logger.MIN_LEVEL is logger.DEBUG:
        warning = "{}:{} {}: {}\n".format(args[1], args[2], args[0].__name__, msg)
    else:
        warning = "{}\n".format(msg)

    return warning


def warn(msg: str, *args) -> None:
    """
    Custom definition of :py:func:`gym.logger.warn` function.
    """

    if logger.MIN_LEVEL <= logger.WARN:
        warnings.warn(colorize('%s: %s' % ('WARN', msg % args), 'yellow'), stacklevel=2)


# Monkey patch warning formatting
warnings.formatwarning = custom_formatwarning


def set_level(level: int, scenario_level: int = None) -> None:
    """
    Set the verbosity level of both :py:mod:`gym` and :py:mod:`gym_ignition`.

    Accepted values:

    - :py:const:`gym.logger.DEBUG` (10)
    - :py:const:`gym.logger.INFO` (20)
    - :py:const:`gym.logger.WARN` (30)
    - :py:const:`gym.logger.ERROR` (40)
    - :py:const:`gym.logger.DISABLED` (50)

    Args:
        level: The desired verbosity level.
        scenario_level: The desired ScenarIO verbosity level (defaults to ``level``).
    """

    # Set the gym verbosity
    logger.set_level(level)

    try:
        from scenario import gazebo as scenario
    except ImportError:
        return

    if scenario_level is None:
        scenario_level = level

    # Set the ScenarI/O verbosity
    if scenario_level <= logger.DEBUG:
        scenario.set_verbosity(scenario.Verbosity_debug)
    elif scenario_level <= logger.INFO:
        scenario.set_verbosity(scenario.Verbosity_info)
    elif scenario_level <= logger.WARN:
        scenario.set_verbosity(scenario.Verbosity_warning)
    elif scenario_level <= logger.ERROR:
        scenario.set_verbosity(scenario.Verbosity_error)
    else:
        scenario.set_verbosity(scenario.Verbosity_suppress_all)


@contextlib.contextmanager
def gym_verbosity(level: int):

    old_level = gym.logger.MIN_LEVEL
    gym.logger.set_level(level=level)
    yield None
    gym.logger.set_level(old_level)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np
from lxml import etree
from pathlib import Path
from enum import auto, Enum
from typing import Dict, List, NamedTuple, Union


class Distribution(Enum):
    Uniform = auto()
    Gaussian = auto()


class Method(Enum):
    Absolute = auto()
    Additive = auto()
    Coefficient = auto()


class RandomizationData(NamedTuple):
    xpath: str
    distribution: str
    parameters: "DistributionParameters"
    method: Method
    ignore_zeros: bool = False
    force_positive: bool = False
    element: etree.Element = None


class GaussianParams(NamedTuple):
    variance: float
    mean: float = None


class UniformParams(NamedTuple):
    low: float
    high: float


DistributionParameters = Union[UniformParams, GaussianParams]


class RandomizationDataBuilder:
    """
    Builder class of a :py:class:`~gym_ignition.randomizers.model.sdf.RandomizationData`
    object.

    Args:
        randomizer: The :py:class:`~gym_ignition.randomizers.model.sdf.SDFRandomizer`
            object to which the created randomization will be inserted.
    """

    def __init__(self, randomizer: "SDFRandomizer"):

        self.storage: Dict = {}
        self.randomizer = randomizer

    def at_xpath(self, xpath: str) -> "RandomizationDataBuilder":
        """
        Set the XPath pattern associated to the randomization.

        Args:
            xpath: The XPath pattern.

        Returns:
            The randomization builder to allow chaining methods.
        """
        self.storage["xpath"] = xpath
        return self

    def sampled_from(self,
                     distribution: Distribution,
                     parameters: DistributionParameters) -> "RandomizationDataBuilder":
        """
        Set the distribution associated to the randomization.

        Args:
            distribution: The desired distribution.
            parameters: The namedtuple with the parameters of the distribution.

        Returns:
            The randomization builder to allow chaining methods.
        """

        self.storage["distribution"] = distribution
        self.storage["parameters"] = parameters

        if self.storage["distribution"] is Distribution.Gaussian and \
                not isinstance(parameters, GaussianParams):
            raise ValueError("Wrong parameters type")

        if self.storage["distribution"] is Distribution.Uniform and \
                not isinstance(parameters, UniformParams):
            raise ValueError("Wrong parameters type")

        return self

    def method(self, method: Method) -> "RandomizationDataBuilder":
        """
        Set the randomization method.

        Args:
            method: The desired randomization method.

        Returns:
            The randomization builder to allow chaining methods.
        """

        self.storage["method"] = method
        return self

    def ignore_zeros(self, ignore_zeros: bool) -> "RandomizationDataBuilder":
        """
        Ignore the randomization of values that are zero.

        If the value to randomize has a default value of 0 in the SDF, when this method
        is chained the randomization is skipped. In the case of a multi-match XPath
        pattern, the values that are not zero are not skipped.

        Args:
            ignore_zeros: True if zeros should be ignored, false otherwise.

        Returns:
            The randomization builder to allow chaining methods.
        """

        self.storage["ignore_zeros"] = ignore_zeros
        return self

    def force_positive(self, force_positive: bool = True) -> "RandomizationDataBuilder":
        """
        Force the randomized value to be greater than zero.

        This option is helpful to enforce that values e.g. the mass will stay positive
        regardless of the applied distribution parameters.

        Args:
            force_positive: True to force positive parameters, false otherwise.

        Returns:
            The randomization builder to allow chaining methods.
        """

        self.storage["force_positive"] = force_positive
        return self

    def add(self) -> None:
        """
        Close the chaining of methods and return to the SDF randomizer the configuration.

        Raises:
            RuntimeError: If the XPath pattern does not find any match in the SDF.
        """

        data = RandomizationData(**self.storage)

        if len(self.randomizer.find_xpath(data.xpath)) == 0:
            raise RuntimeError(f"Failed to find element matching XPath '{data.xpath}'")

        self.randomizer.insert(randomization_data=data)


class SDFRandomizer:
    """
    Randomized SDF files generator.

    Args:
        sdf_model: The absolute path to the SDF file.

    Raises:
       ValueError: If the SDF file does not exist.
    """

    def __init__(self, sdf_model: str):

        self._sdf_file = sdf_model

        if not Path(self._sdf_file).is_file():
            raise ValueError(f"File '{sdf_model}' does not exist")

        # Initialize the root
        tree = self._get_tree_from_file(self._sdf_file)
        self._root: etree.Element = tree.getroot()

        # List of randomizations
        self._randomizations: List[RandomizationData] = []

        # List of default values used with Method.Coefficient
        self._default_values: Dict[etree.Element, float] = {}

        # Store an independent RNG
        self.rng = np.random.default_rng()

    def seed(self, seed: int) -> None:
        """
        Seed the SDF randomizer.

        Args:
            seed: The seed number.
        """
        self.rng = np.random.default_rng(seed)

    def find_xpath(self, xpath: str) -> List[etree.Element]:
        """
        Find the elements that match an XPath pattern.

        This method could be helpful to test the matches of a XPath pattern before using
        it in :py:meth:`~gym_ignition.randomizers.model.sdf.RandomizationDataBuilder.at_xpath`.

        Args:
            xpath: The XPath pattern.

        Return:
            A list of elements matching the XPath pattern.
        """
        return self._root.findall(xpath)

    def process_data(self) -> None:
        """
        Process all the inserted randomizations.

        Raises:
            RuntimeError: If the XPath of a randomization has no matches.
        """

        # Since we support multi-match XPaths, we expand all the individual matches
        expanded_randomizations = []

        for data in self._randomizations:

            # Find all the matches
            elements: List[etree.Element] = self._root.findall(path=data.xpath)

            if len(elements) == 0:
                raise RuntimeError(f"Failed to find elements from XPath '{data.xpath}'")

            for element in elements:

                if data.ignore_zeros and float(self._get_element_text(element)) == 0:
                    continue

                # Get the precise XPath to the element
                element_xpath = element.getroottree().getpath(element)

                # Get the parameters
                params = data.parameters

                if data.method in {Method.Additive, Method.Coefficient}:
                    element_text = float(self._get_element_text(element))
                    self._default_values[element] = element_text

                # Update the data
                complete_data = data._replace(
                    xpath=element_xpath, element=element, parameters=params)

                expanded_randomizations.append(complete_data)

        # Store the updated data
        self._randomizations = expanded_randomizations

    def sample(self, pretty_print=False) -> str:
        """
        Sample a randomized SDF string.

        Args:
            pretty_print: True to pretty print the output.

        Raises:
            ValueError: If the distribution of a randomization is not recognized.
            ValueError: If the method of a randomization is not recognized.

        Returns:
            The randomized model as SDF string.
        """

        for data in self._randomizations:

            if data.distribution is Distribution.Gaussian:

                sample = self.rng.normal(loc=data.parameters.mean,
                                         scale=data.parameters.variance)

            elif data.distribution is Distribution.Uniform:

                sample = self.rng.uniform(low=data.parameters.low,
                                          high=data.parameters.high)

            else:
                raise ValueError("Distribution not recognized")

            if data.force_positive:
                sample = max(sample, 0.0)

            # Update the value
            if data.method is Method.Absolute:

                data.element.text = str(sample)

            elif data.method is Method.Additive:

                default_value = self._default_values[data.element]
                data.element.text = str(sample + default_value)

            elif data.method is Method.Coefficient:

                default_value = self._default_values[data.element]
                data.element.text = str(sample * default_value)

            else:
                raise ValueError("Method not recognized")

        return etree.tostring(self._root, pretty_print=pretty_print).decode()

    def new_randomization(self) -> RandomizationDataBuilder:
        """
        Start the chaining to build a new randomization.

        Return:
            A randomization builder.
        """
        return RandomizationDataBuilder(randomizer=self)

    def insert(self, randomization_data) -> None:
        """
        Insert a randomization.

        Args:
            randomization_data: A new randomization.
        """
        self._randomizations.append(randomization_data)

    def get_active_randomizations(self) -> List[RandomizationData]:
        """
        Return the active randomizations.

        This method could be helpful also in the case of multi-match XPath patterns to
        validate that the inserted randomizations have been processed correctly.

        Returns:
            The list of the active randomizations.
        """
        return self._randomizations

    def clean(self) -> None:
        """
        Clean the SDF randomizer.
        """

        self._randomizations = []
        self._default_values = {}

        tree = self._get_tree_from_file(self._sdf_file)
        self._root = tree.getroot()

    @staticmethod
    def _get_tree_from_file(xml_file) -> etree.ElementTree:

        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse(source=xml_file, parser=parser)

        return tree

    @staticmethod
    def _get_element_text(element: etree.Element) -> str:

        text = element.text

        if text is None:
            raise RuntimeError(f"The element {element.tag} does not have any content")

        return text



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
from typing import List
from os.path import exists, isfile
from gym_ignition.utils import logger

GYM_IGNITION_DATA_PATH = []


def get_search_paths() -> List[str]:
    global GYM_IGNITION_DATA_PATH
    return GYM_IGNITION_DATA_PATH


def add_path(data_path: str) -> None:
    if not exists(data_path):
        logger.warn(f"The path '{data_path}' does not exist. Not added to the data path.")
        return

    global GYM_IGNITION_DATA_PATH

    for path in GYM_IGNITION_DATA_PATH:
        if path == data_path:
            logger.debug(f"The path '{data_path}' is already present in the data path")
            return

    logger.debug(f"Adding new search path: '{data_path}'")
    GYM_IGNITION_DATA_PATH.append(data_path)


def add_path_from_env_var(env_variable: str) -> None:
    if env_variable not in os.environ:
        logger.warn(f"Failed to find '{env_variable}' environment variable")
        return

    # Get the environment variable
    env_var_content = os.environ[env_variable]

    # Remove leading ':' characters
    if env_var_content[0] == ':':
        env_var_content = env_var_content[1:]

    # Split multiple value
    env_var_paths = env_var_content.split(":")

    for path in env_var_paths:
        add_path(path)


def find_resource(file_name: str) -> str:
    file_abs_path = ""
    global GYM_IGNITION_DATA_PATH

    logger.debug(f"Looking for file '{file_name}'")

    # Handle if the path is absolute
    if os.path.isabs(file_name):
        if isfile(file_name):
            logger.debug(f"  Found resource: '{file_name}'")
            return file_name
        else:
            raise FileNotFoundError(f"Failed to find resource '{file_name}'")

    # Handle if the path is relative
    for path in GYM_IGNITION_DATA_PATH:
        logger.debug(f"  Exploring folder '{path}'")
        path_with_slash = path if path[-1] == '/' else path + "/"
        candidate_abs_path = path_with_slash + file_name

        if isfile(candidate_abs_path):
            logger.debug(f"  Found resource: '{candidate_abs_path}'")
            file_abs_path = candidate_abs_path
            break

    if not file_abs_path:
        raise FileNotFoundError(f"Failed to find resource '{file_name}'")

    return file_abs_path



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gym_ignition

import numpy as np
from typing import Tuple
from gym_ignition_environments import models
from scenario import gazebo as scenario_gazebo
from gym_ignition.context.gazebo import controllers
from gym_ignition.rbd.idyntree import inverse_kinematics_nlp
from ..common.utils import default_world_fixture as default_world

# Set the verbosity
scenario_gazebo.set_verbosity(scenario_gazebo.Verbosity_debug)


@pytest.mark.parametrize("default_world", [(1.0 / 1_000, 1.0, 1)], indirect=True)
def test_inverse_kinematics(default_world: Tuple[scenario_gazebo.GazeboSimulator,
                                              scenario_gazebo.World]):

    # Get the simulator and the world
    gazebo, world = default_world

    # Get the robot
    panda = models.panda.Panda(world=world)
    gazebo.run(paused=True)

    # Control all the joints
    controlled_joints = panda.joint_names()

    # Set the controller period
    assert panda.set_controller_period(gazebo.step_size())
    assert panda.controller_period() == gazebo.step_size()

    # Insert the ComputedTorqueFixedBase controller
    assert panda.to_gazebo().insert_model_plugin(*controllers.ComputedTorqueFixedBase(
        kp=[20.0] * panda.dofs(),
        ki=[0.0] * panda.dofs(),
        kd=[5.0] * panda.dofs(),
        urdf=panda.get_model_file(),
        joints=controlled_joints,
    ).args())

    # Create the IK object
    ik = inverse_kinematics_nlp.InverseKinematicsNLP(
        urdf_filename=panda.get_model_file(),
        considered_joints=controlled_joints,
        joint_serialization=panda.joint_names())

    # Initialize IK
    ik.initialize(verbosity=1,
                  floating_base=False,
                  cost_tolerance=1E-8)

    # There should be no active targets
    assert len(ik.get_active_target_names()) == 0

    # Add the cartesian target of the end effector
    end_effector = "end_effector_frame"
    ik.add_target(frame_name=end_effector,
                  target_type=inverse_kinematics_nlp.TargetType.POSE,
                  as_constraint=False)

    assert set(ik.get_active_target_names()) == {end_effector}
    assert ik.get_target_data(target_name=end_effector).type == \
           inverse_kinematics_nlp.TargetType.POSE

    # Desired EE position
    target_ee_position = np.array([0.5, -0.3, 1.0])
    target_ee_quaternion = np.array([0.0, 1.0, 0.0, 0.0])

    # Update the target transform
    ik.update_transform_target(target_name=end_effector,
                               position=target_ee_position,
                               quaternion=target_ee_quaternion)

    # Check that the target transform was stored
    assert isinstance(ik.get_target_data(target_name=end_effector).data,
                      inverse_kinematics_nlp.TransformTargetData)
    assert ik.get_target_data(target_name=end_effector).data.position == \
           pytest.approx(target_ee_position)
    assert ik.get_target_data(target_name=end_effector).data.quaternion == \
           pytest.approx(target_ee_quaternion)

    # Set the current configuration
    ik.set_current_robot_configuration(
        base_position=np.array(panda.base_position()),
        base_quaternion=np.array(panda.base_orientation()),
        joint_configuration=np.array(panda.joint_positions()))

    # Solve the IK problem
    ik.solve()

    # Get the full solution
    ik_solution = ik.get_full_solution()

    # Validate the solution
    assert ik_solution.joint_configuration.size == len(controlled_joints)
    assert ik_solution.base_position == pytest.approx(np.array([0.0, 0, 0]))
    assert ik_solution.base_quaternion == pytest.approx(np.array([1.0, 0, 0, 0]))

    # Set the desired joint configuration
    assert panda.set_joint_position_targets(ik_solution.joint_configuration,
                                            controlled_joints)
    assert panda.set_joint_velocity_targets([0.0] * len(controlled_joints))
    assert panda.set_joint_acceleration_targets([0.0] * len(controlled_joints))

    for _ in range(5_000):
        gazebo.run()

    # Check that the desired joint configuration has been reached
    assert panda.joint_positions(controlled_joints) == \
           pytest.approx(ik_solution.joint_configuration, abs=np.deg2rad(0.1))

    # Get the end effector pose
    ee_position = panda.get_link(link_name=end_effector).position()
    ee_quaternion = panda.get_link(link_name=end_effector).orientation()

    assert ee_position == pytest.approx(target_ee_position, abs=0.005)
    assert ee_quaternion == pytest.approx(target_ee_quaternion, abs=0.005)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym_ignition.base.task
from gym_ignition import randomizers
from scenario import gazebo as scenario


class DART(randomizers.abc.PhysicsRandomizer):
    """
    Class that configures the Gazebo World of a Task to use the DART physics engine.

    Note:
        This class does not apply any physics randomization.
    """

    def __init__(self):

        super().__init__()

    def get_engine(self):

        return scenario.PhysicsEngine_dart

    def randomize_physics(self, task: gym_ignition.base.task.Task, **kwargs) -> None:

        pass



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from pathlib import Path
from setuptools import setup
from cmake_build_extension import BuildExtension, CMakeExtension

# Read the contents of the README file
this_directory = Path(__file__).absolute().parent
with open(this_directory / "README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="scenar-io",  # temporary name waiting for 'scenario'
    author="Diego Ferigo",
    author_email="diego.ferigo@iit.it",
    description="SCENe interfAces for Robot Input/Output.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/robotology/gym-ignition/tree/master/scenario",
    project_urls={
        "Bug Tracker": "https://github.com/robotology/gym-ignition/issues",
        "Documentation": "https://robotology.github.io/gym-ignition",
        "Source Code": "https://github.com/robotology/gym-ignition/tree/master/scenario",
    },
    keywords=["robotics", "gazebo", "ignition", "simulation", "physics", "multibody",
              "dynamics", "physics simulation", "middleware", "real-time"],
    license="LGPL",
    platforms="any",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: POSIX :: Linux",
        "Topic :: Games/Entertainment :: Simulation",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development",
        "Framework :: Robot Framework",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    ],
    use_scm_version=dict(local_scheme="dirty-tag", root="..", relative_to=__file__),
    setup_requires=[
        "wheel",
        "setuptools_scm",
        "cmake-build-extension",
        "idyntree>=3.1",
    ],
    python_requires=">=3.0",
    ext_modules=[CMakeExtension(name="ScenarioCMakeProject",
                                install_prefix="scenario",
                                cmake_build_type="PyPI",
                                cmake_depends_on=["idyntree"],
                                disable_editable=True)],
    cmdclass=dict(build_ext=BuildExtension),
    zip_safe=False,
)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import List
from scenario import core as scenario
from gym_ignition.utils.scenario import get_unique_model_name
from gym_ignition.scenario import model_wrapper, model_with_file


class Panda(model_wrapper.ModelWrapper,
            model_with_file.ModelWithFile):

    def __init__(self,
                 world: scenario.World,
                 position: List[float] = (0.0, 0.0, 0.0),
                 orientation: List[float] = (1.0, 0, 0, 0),
                 model_file: str = None):

        # Get a unique model name
        model_name = get_unique_model_name(world, "panda")

        # Initial pose
        initial_pose = scenario.Pose(position, orientation)

        # Get the default model description (URDF or SDF) allowing to pass a custom model
        if model_file is None:
            model_file = Panda.get_model_file()

        # Insert the model
        ok_model = world.to_gazebo().insert_model(model_file,
                                                  initial_pose,
                                                  model_name)

        if not ok_model:
            raise RuntimeError("Failed to insert model")

        # Get the model
        model = world.get_model(model_name)

        # Initial joint configuration
        model.to_gazebo().reset_joint_positions(
            [0, -0.785,0, -2.356, 0, 1.571, 0.785],
            [name for name in model.joint_names() if "panda_joint" in name])

        # From:
        # https://github.com/mkrizmancic/franka_gazebo/blob/master/config/default.yaml
        pid_gains_1000hz = {
            'panda_joint1': scenario.PID(50,    0,  20),
            'panda_joint2': scenario.PID(10000, 0, 500),
            'panda_joint3': scenario.PID(100,   0,  10),
            'panda_joint4': scenario.PID(1000,  0,  50),
            'panda_joint5': scenario.PID(100,   0,  10),
            'panda_joint6': scenario.PID(100,   0,  10),
            'panda_joint7': scenario.PID(10,  0.5, 0.1),
            'panda_finger_joint1': scenario.PID(100, 0, 50),
            'panda_finger_joint2': scenario.PID(100, 0, 50),
        }

        # Check that all joints have gains
        if not set(model.joint_names()) == set(pid_gains_1000hz.keys()):
            raise ValueError("The number of PIDs does not match the number of joints")

        # Set the PID gains
        for joint_name, pid in pid_gains_1000hz.items():

            if not model.get_joint(joint_name).set_pid(pid=pid):
                raise RuntimeError(f"Failed to set the PID of joint '{joint_name}'")

        # Set the default PID update period
        assert model.set_controller_period(1000.0)

        # Initialize base class
        super().__init__(model=model)

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models
        return gym_ignition_models.get_model_file("panda")



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from pathlib import Path
from setuptools import setup, find_packages

# Read the contents of the README file
this_directory = Path(__file__).absolute().parent
with open(this_directory / "README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="gym-ignition",
    author="Diego Ferigo",
    author_email="diego.ferigo@iit.it",
    description="Gym-Ignition: A toolkit for developing OpenAI Gym environments "
                "simulated with Ignition Gazebo.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/robotology/gym-ignition",
    project_urls={
        "Bug Tracker": "https://github.com/robotology/gym-ignition/issues",
        "Documentation": "https://robotology.github.io/gym-ignition",
        "Source Code": "https://github.com/robotology/gym-ignition",
    },
    keywords=["openai", "gym", "reinforcement learning", "rl", "environment", "gazebo",
              "robotics", "ignition", "humanoid", "panda", "icub", "urdf", "sdf"],
    license="LGPL",
    platforms="any",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: POSIX :: Linux",
        "Topic :: Games/Entertainment :: Simulation",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development",
        "Framework :: Robot Framework",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    ],
    use_scm_version=dict(local_scheme="dirty-tag"),
    setup_requires=[
        "wheel",
        "setuptools_scm",
        "cmake-build-extension",
    ],
    python_requires=">=3.8",
    install_requires=[
        "scenar-io",  # temporary name waiting for 'scenario'
        "gym>=0.13.1",
        "numpy",
        "gym_ignition_models",
        "lxml",
        "idyntree",
    ],
    extras_require=dict(
        website=[
            "sphinx",
            "sphinx-book-theme",
            "sphinx-autodoc-typehints",
            "sphinx_fontawesome",
            "sphinx-multiversion",
            "sphinx-tabs",
            "breathe",
        ],
        test=[
            "pytest",
            "pytest-xvfb",
            "pytest-icdiff",
        ]
    ),
    packages=find_packages("python"),
    package_dir={"": "python"},
    zip_safe=False,
)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
import gym
import numpy as np
from typing import Tuple
from gym_ignition.base import task
from scenario import core as scenario
from gym_ignition.utils.typing import Action, Reward, Observation
from gym_ignition.utils.typing import ActionSpace, ObservationSpace


class CartPoleContinuousBalancing(task.Task, abc.ABC):

    def __init__(self,
                 agent_rate: float,
                 reward_cart_at_center: bool = True,
                 **kwargs):

        # Initialize the Task base class
        task.Task.__init__(self, agent_rate=agent_rate)

        # Name of the cartpole model
        self.model_name = None

        # Space for resetting the task
        self.reset_space = None

        # Private attributes
        self._reward_cart_at_center = reward_cart_at_center

        # Variables limits
        self._x_threshold = 2.4  # m
        self._dx_threshold = 20.0  # m /s
        self._q_threshold = np.deg2rad(12)  # rad
        self._dq_threshold = np.deg2rad(3 * 360)  # rad / s

    def create_spaces(self) -> Tuple[ActionSpace, ObservationSpace]:

        # Create the action space
        max_force = 50.0  # Nm
        action_space = gym.spaces.Box(low=np.array([-max_force]),
                                      high=np.array([max_force]),
                                      dtype=np.float32)

        # Configure reset limits
        high = np.array([
            self._x_threshold,   # x
            self._dx_threshold,  # dx
            self._q_threshold,   # q
            self._dq_threshold   # dq
        ])

        # Configure the reset space
        self.reset_space = gym.spaces.Box(low=-high,
                                          high=high,
                                          dtype=np.float32)

        # Configure the observation space
        obs_high = high.copy() * 1.2
        observation_space = gym.spaces.Box(low=-obs_high,
                                           high=obs_high,
                                           dtype=np.float32)

        return action_space, observation_space

    def set_action(self, action: Action) -> None:

        # Get the force value
        force = action.tolist()[0]

        # Set the force value
        model = self.world.get_model(self.model_name)
        ok_force = model.get_joint("linear").set_generalized_force_target(force)

        if not ok_force:
            raise RuntimeError("Failed to set the force to the cart")

    def get_observation(self) -> Observation:

        # Get the model
        model = self.world.get_model(self.model_name)

        # Get the new joint positions and velocities
        q, x = model.joint_positions(["pivot", "linear"])
        dq, dx = model.joint_velocities(["pivot", "linear"])

        # Create the observation
        observation = Observation(np.array([x, dx, q, dq]))

        # Return the observation
        return observation

    def get_reward(self) -> Reward:

        # Calculate the reward
        reward = 1.0 if not self.is_done() else 0.0

        if self._reward_cart_at_center:

            # Get the observation
            x, dx, _, _ = self.get_observation()

            reward = reward \
                - 0.10 * np.abs(x) \
                - 0.10 * np.abs(dx) \
                - 10.0 * (x >= self._x_threshold)

        return reward

    def is_done(self) -> bool:

        # Get the observation
        observation = self.get_observation()

        # The environment is done if the observation is outside its space
        done = not self.reset_space.contains(observation)

        return done

    def reset_task(self) -> None:

        if self.model_name not in self.world.model_names():
            raise RuntimeError("Cartpole model not found in the world")

        # Get the model
        model = self.world.get_model(self.model_name)

        # Control the cart in force mode
        linear = model.get_joint("linear")
        ok_control_mode = linear.set_control_mode(scenario.JointControlMode_force)

        if not ok_control_mode:
            raise RuntimeError("Failed to change the control mode of the cartpole")

        # Create a new cartpole state
        x, dx, q, dq = self.np_random.uniform(low=-0.05, high=0.05, size=(4,))

        # Reset the cartpole state
        ok_reset_pos = model.to_gazebo().reset_joint_positions([x, q], ["linear", "pivot"])
        ok_reset_vel = model.to_gazebo().reset_joint_velocities([dx, dq], ["linear", "pivot"])

        if not (ok_reset_pos and ok_reset_vel):
            raise RuntimeError("Failed to reset the cartpole state")



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

from scenario import core
from ..common import utils
from scenario import gazebo as scenario
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_load_default_world(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    world = gazebo.get_world()

    assert world.id() != 0
    assert world.time() == 0.0
    assert world.name() == "default"


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_load_default_world_from_file(gazebo: scenario.GazeboSimulator):

    empty_world_sdf = utils.get_empty_world_sdf()

    assert gazebo.insert_world_from_sdf(empty_world_sdf)

    assert gazebo.initialize()
    world = gazebo.get_world()

    assert world.id() != 0
    assert world.time() == 0.0
    assert world.name() == "default"


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_rename_default_world(gazebo: scenario.GazeboSimulator):

    empty_world_sdf = utils.get_empty_world_sdf()
    assert gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld")

    assert gazebo.initialize()
    world = gazebo.get_world()

    assert world.id() != 0
    assert world.time() == 0.0
    assert world.name() == "myWorld"

    world1 = gazebo.get_world("myWorld")

    assert world1.id() == world.id()
    assert world1.time() == 0.0
    assert world1.name() == "myWorld"


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_world_api(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    world = gazebo.get_world()

    gravity = [0, 0, 10.0]
    assert world.set_gravity(gravity)
    assert world.gravity() == pytest.approx(gravity)

    assert len(world.model_names()) == 0

    # Insert a model from an non existent file
    assert not world.insert_model("")

    # Insert first cube with default name and pose
    cube_urdf = utils.get_cube_urdf()
    assert world.insert_model(cube_urdf)
    assert len(world.model_names()) == 1

    default_model_name = scenario.get_model_name_from_sdf(cube_urdf)
    assert default_model_name in world.model_names()
    cube1 = world.get_model(default_model_name)
    # TODO: assert cube1

    assert cube1.name() == default_model_name
    assert cube1.base_position() == pytest.approx([0, 0, 0])
    assert cube1.base_orientation() == pytest.approx([1, 0, 0, 0])

    # Inserting a model with the same name should fail
    assert not world.insert_model(cube_urdf)
    assert len(world.model_names()) == 1

    # Insert second cube with custom name and pose
    custom_model_name = "other_cube"
    assert custom_model_name != default_model_name
    custom_model_pose = core.Pose([1, 1, 0], [0, 0, 0, 1])

    assert world.insert_model(cube_urdf, custom_model_pose, custom_model_name)
    assert custom_model_name in world.model_names()
    assert len(world.model_names()) == 2

    cube2 = world.get_model(custom_model_name)
    assert cube1 != cube2
    # TODO: assert cube2

    assert cube2.name() == custom_model_name
    assert cube2.base_position() == pytest.approx(custom_model_pose.position)
    assert cube2.base_orientation() == pytest.approx(custom_model_pose.orientation)

    # insert a cube from urdf string
    cube_3_pose = core.Pose([1, 0, 0], [0, 0, 0, 1])
    assert world.insert_model_from_string(
        utils.get_cube_urdf_string(), cube_3_pose, "cube3"
    )
    assert "cube3" in world.model_names()

    # insert a cube from sdf string
    cube_4_pose = core.Pose([2, 0, 0], [0, 0, 0, 1])
    assert world.insert_model_from_string(
        utils.get_cube_sdf_string(), cube_4_pose, "cube4"
    )
    assert "cube4" in world.model_names()

    # Remove the first model (requires either a paused or unpaused step)
    assert world.remove_model(default_model_name)
    assert len(world.model_names()) == 4
    gazebo.run(paused=True)
    assert len(world.model_names()) == 3

    # Without the physics system, the time should not increase
    gazebo.run()
    gazebo.run()
    gazebo.run()
    assert world.time() == 0.0


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_world_physics_plugin(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    world = gazebo.get_world()

    dt = gazebo.step_size()

    assert world.time() == 0

    # Insert a cube
    cube_urdf = utils.get_cube_urdf()
    cube_name = "my_cube"
    cube_pose = core.Pose([0, 0, 1.0], [1, 0, 0, 0])
    assert world.insert_model(cube_urdf, cube_pose, cube_name)
    assert cube_name in world.model_names()

    cube = world.get_model(cube_name)

    # There's no physics, the cube should not move
    for _ in range(10):
        gazebo.run()
        assert cube.base_position() == cube_pose.position

    assert world.time() == 0

    # Insert the Physics system
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    # After the first step, the physics catches up with time
    gazebo.run()
    assert world.time() == pytest.approx(11 * dt)

    # The cube should start falling
    assert cube.base_position()[2] < cube_pose.position[2]

    gazebo.run()
    assert world.time() == pytest.approx(12 * dt)

    # Paused steps do not increase the time
    gazebo.run(paused=True)
    assert world.time() == pytest.approx(12 * dt)


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1), (1e-9, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_sim_time_starts_from_zero(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    world = gazebo.get_world()

    dt = gazebo.step_size()

    assert world.time() == 0
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)
    assert world.time() == 0

    gazebo.run(paused=True)
    assert world.time() == 0

    gazebo.run(paused=False)
    assert world.time() == dt

    gazebo.run(paused=False)
    assert world.time() == 2 * dt

    gazebo.run(paused=False)
    assert world.time() == pytest.approx(3 * dt)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------
from datetime import datetime
project = 'gym-ignition'
copyright = f'{datetime.now().year}, Istituto Italiano di Tecnologia'
author = 'Diego Ferigo'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    "sphinx.ext.extlinks",
    'sphinx_autodoc_typehints',
    "sphinx_multiversion",
    "sphinx_fontawesome",
    'breathe',
    'sphinx_tabs.tabs',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "repository_url": "https://github.com/robotology/gym-ignition",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "path_to_docs": "docs/sphinx",
    "home_page_in_toc": True,
    "use_download_button": False,
    "use_fullscreen_button": True,
    "single_page": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'pastie'


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for breathe extension ----------------------------------------------

breathe_default_project = "scenario"

# -- Options for sphinx_multiversion extension ----------------------------------

# From: https://holzhaus.github.io/sphinx-multiversion
smv_prefer_remote_refs = False
smv_remote_whitelist = r'^(origin|upstream)$'
smv_tag_whitelist = r'^dummy'
smv_branch_whitelist = r'^(master|devel|docs/.*)$'
smv_released_pattern = r'^tags/.*$'
smv_outputdir_format = '{ref.name}'

html_sidebars = {
    "**": [
        "sbt-sidebar-nav.html",
        "versioning.html",
        "search-field.html",
        "sbt-sidebar-footer.html",
    ]
}

# -- Options for extlinks extension ----------------------------------
extlinks = {
    'issue': ('https://github.com/robotology/gym-ignition/issues/%s', '#'),
    'pr': ('https://github.com/robotology/gym-ignition/pull/%s', '#'),
}
